# Mappings from Browser names we use to known client hint values
# fmt: off
APP_TO_CLIENT_HINT_MAPPING = {
    'Chrome'                    : ['Google Chrome'],
    'Chrome Webview'            : ['Android WebView'],
    'DuckDuckGo Privacy Browser': ['DuckDuckGo'],
    'Edge WebView'              : ['Microsoft Edge WebView2'],
    'Mi Browser'                : ['Miui Browser', 'XiaoMiBrowser'],
    'Microsoft Edge'            : ['Edge'],
    'Norton Private Browser'    : ['Norton Secure Browser'],
    'Opera GX'                  : ['Opera GX Android'],
    'Opera Mini'                : ['Opera Mini Android'],
    'Vewd Browser'              : ['Vewd Core'],
    'Yandex Browser'            : ['YaSearchBrowser'],
}

CLIENT_HINT_TO_APP_MAP = {}
for dd_app_name, ch_app_names in APP_TO_CLIENT_HINT_MAPPING.items():
    for ch_app_name in ch_app_names:
        CLIENT_HINT_TO_APP_MAP[ch_app_name.lower()] = dd_app_name
        CLIENT_HINT_TO_APP_MAP[ch_app_name] = dd_app_name


# Mappings from OS names we use to known client hint values
OS_CLIENT_HINT_MAPPING = {
    'GNU/Linux': ['Linux'],
    'Mac'      : ['MacOS', 'macOS'],
}

# Operating system families that are known as desktop only
DESKTOP_OS = {
    'AmigaOS', 'IBM', 'GNU/Linux', 'Mac', 'Unix', 'Windows', 'BeOS', 'Chrome OS', 'Chromium OS',
    'OpenVMS',
}

# Sub-list of Desktop OS that never falsely appear in UAs
ALWAYS_DESKTOP_OS = {
    'AmigaOS', 'IBM', 'Unix', 'BeOS', 'Arch Linux',
}

APPLE_OS_NAMES = {
    'tvOS',
    'watchOS',
    'iPadOS',
    'iOS',
    'macOS',
    'Mac',
}

# Fire OS version mapping
FIREOS_VERSION_MAPPING = {
    '14'   : '14',
    '13'   : '14',
    '12'   : '14',
    '11'   : '8',
    '10'   : '8',
    '9'    : '7',
    '7'    : '6',
    '5'    : '5',
    '4.4.3': '4.5.1',
    '4.4.2': '4',
    '4.2.2': '3',
    '4.0.3': '3',
    '4.0.2': '3',
    '4'    : '2',
    '2'    : '1',
}

# Lineage OS version mapping
LINEAGEOS_VERSION_MAPPING = {
    '16'   : '23',
    '15'   : '22',
    '14'   : '21',
    '13'   : '20.0',
    '12.1' : '19.1',
    '12'   : '19.0',
    '11'   : '18.0',
    '10'   : '17.0',
    '9'    : '16.0',
    '8.1.0': '15.1',
    '8.0.0': '15.0',
    '7.1.2': '14.1',
    '7.1.1': '14.1',
    '7.0'  : '14.0',
    '6.0.1': '13.0',
    '6.0'  : '13.0',
    '5.1.1': '12.1',
    '5.0.2': '12.0',
    '5.0'  : '12.0',
    '4.4.4': '11.0',
    '4.3'  : '10.2',
    '4.2.2': '10.1',
    '4.0.4': '9.1.0',
}
# fmt: off

ANDROID_APPS = {
    'com.hisense.odinbrowser',
    'com.seraphic.openinet.pre',
    'com.appssppa.idesktoppcbrowser',
    'every.browser.inc',
}


OPERATING_SYSTEMS = {
    'AIX': 'AIX',
    'AND': 'Android',
    'ADR': 'Android TV',
    'ALP': 'Alpine Linux',
    'AMZ': 'Amazon Linux',
    'AMG': 'AmigaOS',
    'ARM': 'Armadillo OS',
    'ARO': 'AROS',
    'ATV': 'tvOS',
    'ARC': 'ArcaOS',
    'ARL': 'Arch Linux',
    'AOS': 'AOSC OS',
    'ASP': 'ASPLinux',
    'AZU': 'Azure Linux',
    'BTR': 'BackTrack',
    'SBA': 'Bada',
    'BYI': 'Baidu Yi',
    'BEO': 'BeOS',
    'BLB': 'BlackBerry OS',
    'QNX': 'BlackBerry Tablet OS',
    'PAN': 'blackPanther OS',
    'BOS': 'Bliss OS',
    'BMP': 'Brew',
    'BSN': 'BrightSignOS',
    'BS1': 'BSD',
    'CAI': 'Caixa Mágica',
    'CES': 'CentOS',
    'CST': 'CentOS Stream',
    'CLO': 'Clear Linux OS',
    'CLR': 'ClearOS Mobile',
    'COS': 'Chrome OS',
    'CRS': 'Chromium OS',
    'CHN': 'China OS',
    'CON': 'Contiki',
    'COL': 'Coolita OS',
    'CYN': 'CyanogenMod',
    'DEB': 'Debian',
    'DEE': 'Deepin',
    'DFB': 'DragonFly',
    'DVK': 'DVKBuntu',
    'ELE': 'ElectroBSD',
    'ELM': 'elementary OS',
    'EUL': 'EulerOS',
    'FED': 'Fedora',
    'FEN': 'Fenix',
    'FOS': 'Firefox OS',
    'FIR': 'Fire OS',
    'FOR': 'Foresight Linux',
    'FRE': 'Freebox',
    'BSD': 'FreeBSD',
    'FRI': 'FRITZ!OS',
    'FYD': 'FydeOS',
    'FUC': 'Fuchsia',
    'GNT': 'Gentoo',
    'GNX': 'GENIX',
    'GEO': 'GEOS',
    'GHO': 'GhostBSD',
    'GNS': 'gNewSense',
    'GRI': 'GridOS',
    'GTV': 'Google TV',
    'HPX': 'HP-UX',
    'HAI': 'Haiku OS',
    'IPA': 'iPadOS',
    'HAR': 'HarmonyOS',
    'HAS': 'HasCodingOS',
    'HEL': 'HELIX OS',
    'IRI': 'IRIX',
    'INF': 'Inferno',
    'JME': 'Java ME',
    'JOL': 'Joli OS',
    'KOS': 'KaiOS',
    'KAL': 'Kali',
    'KAN': 'Kanotix',
    'KIN': 'KIN OS',
    'KOL': 'KolibriOS',
    'KNO': 'Knoppix',
    'KTV': 'KreaTV',
    'KBT': 'Kubuntu',
    'LIN': 'GNU/Linux',
    'LEA': 'LeafOS',
    'LND': 'LindowsOS',
    'LNS': 'Linspire',
    'LEN': 'Lineage OS',
    'LPU': 'Linpus',
    'LIR': 'Liri OS',
    'LOO': 'Loongnix',
    'LBT': 'Lubuntu',
    'LOS': 'Lumin OS',
    'LUN': 'LuneOS',
    'VLN': 'VectorLinux',
    'MAC': 'Mac',
    'MAE': 'Maemo',
    'MAG': 'Mageia',
    'MDR': 'Mandriva',
    'SMG': 'MeeGo',
    'MET': 'Meta Horizon',
    'MNX': 'MINIX',
    'MOS': 'Mocor OS',
    'MCD': 'MocorDroid',
    'MON': 'moonOS',
    'EZX': 'Motorola EZX',
    'MIN': 'Mint',
    'MLD': 'MildWild',
    'MOR': 'MorphOS',
    'NBS': 'NetBSD',
    'MTK': 'MTK / Nucleus',
    'MRE': 'MRE',
    'NXT': 'NeXTSTEP',
    'NWS': 'NEWS-OS',
    'WII': 'Nintendo',
    'NDS': 'Nintendo Mobile',
    'NOV': 'Nova',
    'NTX': 'NuttX',
    'OS2': 'OS/2',
    'T64': 'OSF1',
    'OSS': 'openSUSE',
    'OBS': 'OpenBSD',
    'OHS': 'OpenHarmony',
    'OVS': 'OpenVMS',
    'OVZ': 'OpenVZ',
    'OWR': 'OpenWrt',
    'OTV': 'Opera TV',
    'ORA': 'Oracle Linux',
    'ORD': 'Ordissimo',
    'ORS': 'Orsay',
    'PAR': 'Pardus',
    'PCL': 'PCLinuxOS',
    'PIC': 'PICO OS',
    'PLN': 'Plan 9',
    'PLA': 'Plasma Mobile',
    'PSP': 'PlayStation Portable',
    'PS3': 'PlayStation',
    'PVE': 'Proxmox VE',
    'PUF': 'Puffin OS',
    'PUR': 'PureOS',
    'QTP': 'Qtopia',
    'PIO': 'Raspberry Pi OS',
    'RAS': 'Raspbian',
    'RHT': 'Red Hat',
    'RST': 'Red Star',
    'RED': 'RedOS',
    'REV': 'Revenge OS',
    'RIS': 'risingOS',
    'ROS': 'RISC OS',
    'ROC': 'Rocky Linux',
    'ROK': 'Roku OS',
    'RSO': 'Rosa',
    'ROU': 'RouterOS',
    'REM': 'Remix OS',
    'RRS': 'Resurrection Remix OS',
    'REX': 'REX',
    'RZD': 'RazoDroiD',
    'RXT': 'RTOS & Next',
    'SAB': 'Sabayon',
    'SSE': 'SUSE',
    'SAF': 'Sailfish OS',
    'SCI': 'Scientific Linux',
    'SEE': 'SeewoOS',
    'SER': 'SerenityOS',
    'SIR': 'Sirin OS',
    'SLW': 'Slackware',
    'SMA': 'Smartisan OS',
    'SOS': 'Solaris',
    'SBL': 'Star-Blade OS',
    'SYL': 'Syllable',
    'SYM': 'Symbian',
    'SYS': 'Symbian OS',
    'S40': 'Symbian OS Series 40',
    'S60': 'Symbian OS Series 60',
    'SY3': 'Symbian^3',
    'TEN': 'TencentOS',
    'TDX': 'ThreadX',
    'TIT': 'Titan OS',
    'TIZ': 'Tizen',
    'TIV': 'TiVo OS',
    'TOS': 'TmaxOS',
    'TUR': 'Turbolinux',
    'UBT': 'Ubuntu',
    'ULT': 'ULTRIX',
    'UOS': 'UOS',
    'VID': 'VIDAA',
    'VIZ': 'ViziOS',
    'WAS': 'watchOS',
    'WER': 'Wear OS',
    'WTV': 'WebTV',
    'WHS': 'Whale OS',
    'WIN': 'Windows',
    'WCE': 'Windows CE',
    'WIO': 'Windows IoT',
    'WMO': 'Windows Mobile',
    'WPH': 'Windows Phone',
    'WRT': 'Windows RT',
    'WPO': 'WoPhone',
    'XBX': 'Xbox',
    'XBT': 'Xubuntu',
    'YNS': 'YunOS',
    'ZEN': 'Zenwalk',
    'ZOR': 'ZorinOS',
    'IOS': 'iOS',
    'POS': 'palmOS',
    'WEB': 'Webian',
    'WOS': 'webOS',
    'UNK': 'Unknown',
}

# flip Abbrev / OS for fast membership testing
OS_TO_ABBREV = {os.lower(): abbrev for abbrev, os in OPERATING_SYSTEMS.items()}

# fmt: off
OS_FAMILIES = {
    'Android': [
        'AND', 'CYN', 'FIR', 'REM', 'RZD', 'MLD', 'MCD', 'YNS', 'GRI', 'HAR',
        'ADR', 'CLR', 'BOS', 'REV', 'LEN', 'SIR', 'RRS', 'WER', 'PIC', 'ARM',
        'HEL', 'BYI', 'RIS', 'PUF', 'LEA', 'MET', 'OHS', 'SMA',
    ],
    'AmigaOS': ('AMG', 'MOR', 'ARO'),
    'BlackBerry': ('BLB', 'QNX'),
    'Brew': ('BMP',),
    'BeOS': ('BEO', 'HAI'),
    'Chrome OS': ('COS', 'CRS', 'FYD', 'SEE'),
    'Firefox OS': ('FOS', 'KOS'),
    'Gaming Console': ('WII', 'PS3'),
    'Google TV': ('GTV',),
    'IBM': ('OS2', 'ARC'),
    'iOS': ('IOS', 'ATV', 'WAS', 'IPA'),
    'RISC OS': ('ROS',),
    'GNU/Linux': (
        'LIN', 'ARL', 'DEB', 'KNO', 'MIN', 'UBT', 'KBT', 'XBT', 'LBT', 'FED',
        'RHT', 'VLN', 'MDR', 'GNT', 'SAB', 'SLW', 'SSE', 'CES', 'BTR', 'SAF',
        'ORD', 'TOS', 'RSO', 'DEE', 'FRE', 'MAG', 'FEN', 'CAI', 'PCL', 'HAS',
        'LOS', 'DVK', 'ROK', 'OWR', 'OTV', 'KTV', 'PUR', 'PLA', 'FUC', 'PAR',
        'FOR', 'MON', 'KAN', 'ZEN', 'LND', 'LNS', 'CHN', 'AMZ', 'TEN', 'CST',
        'NOV', 'ROU', 'ZOR', 'RED', 'KAL', 'ORA', 'VID', 'TIV', 'BSN', 'RAS',
        'UOS', 'PIO', 'FRI', 'LIR', 'WEB', 'SER', 'ASP', 'AOS', 'LOO', 'EUL',
        'SCI', 'ALP', 'CLO', 'ROC', 'OVZ', 'PVE', 'RST', 'EZX', 'GNS', 'JOL',
        'TUR', 'QTP', 'WPO', 'PAN', 'VIZ', 'AZU', 'COL', 'OSS', 'ELM', 'LPU',
    ),
    'Mac': ('MAC',),
    'Mobile Gaming Console': ('PSP', 'NDS', 'XBX'),
    'OpenVMS': ('OVS',),
    'Other Mobile': ('WOS', 'POS', 'SBA', 'TIZ', 'SMG', 'MAE', 'LUN', 'GEO', 'CON'),
    'Real-time OS': ('MTK', 'TDX', 'MRE', 'JME', 'REX', 'RXT', 'KOL', 'MOS', 'NTX'),
    'Symbian': ('SYM', 'SYS', 'SY3', 'S60', 'S40'),
    'Unix': (
        'SOS', 'AIX', 'HPX', 'BSD', 'NBS', 'OBS', 'DFB', 'SYL', 'IRI', 'T64',
        'INF', 'ELE', 'GNX', 'ULT', 'NWS', 'NXT', 'SBL', 'BS1', 'GHO', 'PLN',
        'MNX',
    ),
    'WebTV': ('WTV',),
    'Windows': ('WIN',),
    'Windows Mobile': ('WPH', 'WMO', 'WCE', 'WRT', 'WIO', 'KIN'),
    'Other Smart TV': ('WHS', 'TIT', 'ORS'),
    'Unknown': ('UNK',),
}
# fmt: on

FAMILY_FROM_OS = {}
for os, families in OS_FAMILIES.items():
    for family in families:
        FAMILY_FROM_OS[family] = os

for os, families in OS_CLIENT_HINT_MAPPING.items():
    for family in families:
        FAMILY_FROM_OS[family] = os
        FAMILY_FROM_OS[family.lower()] = os

AVAILABLE_ENGINES = {
    'WebKit',
    'Blink',
    'Trident',
    'Text-based',
    'Dillo',
    'iCab',
    'Elektra',
    'Presto',
    'Clecko',
    'Gecko',
    'KHTML',
    'NetFront',
    'Edge',
    'NetSurf',
    'Servo',
    'Goanna',
    'EkiohFlow',
    'Arachne',
    'LibWeb',
    'Maple',
}
AVAILABLE_ENGINES_LOWER_CASE = {engine.lower(): engine for engine in AVAILABLE_ENGINES}

AVAILABLE_BROWSERS = {
    'V1': 'Via',
    '1P': 'Pure Mini Browser',
    '4P': 'Pure Lite Browser',
    '1R': 'Raise Fast Browser',
    'R1': 'Rabbit Private Browser',
    'FQ': 'Fast Browser UC Lite',
    'FJ': 'Fast Explorer',
    '1L': 'Lightning Browser',
    '1C': 'Cake Browser',
    '1I': 'IE Browser Fast',
    '1V': 'Vegas Browser',
    '1O': 'OH Browser',
    '3O': 'OH Private Browser',
    '1X': 'XBrowser Mini',
    '1S': 'Sharkee Browser',
    '2L': 'Lark Browser',
    '3P': 'Pluma',
    '1A': 'Anka Browser',
    'AZ': 'Azka Browser',
    '1D': 'Dragon Browser',
    '1E': 'Easy Browser',
    'DW': 'Dark Web Browser',
    'D6': 'Dark Browser',
    '18': '18+ Privacy Browser',
    '1B': '115 Browser',
    'DM': '1DM Browser',
    '1M': '1DM+ Browser',
    '2B': '2345 Browser',
    '3B': '360 Secure Browser',
    '36': '360 Phone Browser',
    '7B': '7654 Browser',
    'AA': 'Avant Browser',
    'AB': 'ABrowse',
    '00': 'Ace',
    '4A': 'Acoo Browser',
    'BW': 'AdBlock Browser',
    'A7': 'Adult Browser',
    '8A': 'Ai Browser',
    'A9': 'Airfind Secure Browser',
    'AF': 'ANT Fresco',
    'AG': 'ANTGalio',
    'AL': 'Aloha Browser',
    'AH': 'Aloha Browser Lite',
    'A8': 'ALVA',
    '9A': 'AltiBrowser',
    'AM': 'Amaya',
    'A3': 'Amaze Browser',
    'A5': 'Amerigo',
    'AO': 'Amigo',
    'AN': 'Android Browser',
    'AY': 'Anonyv Browser',
    '3A': 'AOL Explorer',
    'AE': 'AOL Desktop',
    'AD': 'AOL Shield',
    'A4': 'AOL Shield Pro',
    '2A': 'Aplix',
    'A6': 'AppBrowzer',
    '0A': 'AppTec Secure Browser',
    'AP': 'APUS Browser',
    'AR': 'Arora',
    'AX': 'Arctic Fox',
    'AV': 'Amiga Voyager',
    'AW': 'Amiga Aweb',
    'PN': 'APN Browser',
    '6A': 'Arachne',
    'RA': 'Arc Search',
    'R5': 'Armorfly Browser',
    'AQ': 'ArtisBrowser',
    'AI': 'Arvin',
    'AK': 'Ask.com',
    'KA': 'Ask Browser',
    'AU': 'Asus Browser',
    'A0': 'Atom',
    'AT': 'Atomic Web Browser',
    'A2': 'Atlas',
    'AS': 'Avast Secure Browser',
    'VG': 'AVG Secure Browser',
    'AC': 'Avira Secure Browser',
    'A1': 'AwoX',
    '7A': 'Awesomium',
    '5B': 'Basic Web Browser',
    'BA': 'Beaker Browser',
    'BM': 'Beamrise',
    'F7': 'BF Browser',
    'BB': 'BlackBerry Browser',
    '6B': 'Bluefy',
    'H1': 'BrowseHere',
    'B8': 'Browser Hup Pro',
    'BD': 'Baidu Browser',
    'BS': 'Baidu Spark',
    'BG': 'Bang',
    'B9': 'Bangla Browser',
    'BI': 'Basilisk',
    'BV': 'Belva Browser',
    'B5': 'Beyond Private Browser',
    'BE': 'Beonex',
    'B2': 'Berry Browser',
    'BT': 'Bitchute Browser',
    '9B': 'BizBrowser',
    'BC': 'Black Browser',
    'BH': 'BlackHawk',
    'Z1': 'Blazer',
    'B0': 'Bloket',
    'BJ': 'Bunjalloo',
    'BL': 'B-Line',
    'B6': 'Black Lion Browser',
    'BU': 'Blue Browser',
    'BO': 'Bonsai',
    'BN': 'Borealis Navigator',
    'BR': 'Brave',
    'BK': 'BriskBard',
    'K2': 'BroKeep Browser',
    'B3': 'Browspeed Browser',
    'BX': 'BrowseX',
    'BZ': 'Browzar',
    'B7': 'Browlser',
    'BQ': 'Browse Safe',
    'M9': 'Browser Mini',
    '4B': 'BrowsBit',
    'BY': 'Biyubi',
    'BF': 'Byffox',
    'B4': 'BXE Browser',
    'CA': 'Camino',
    '5C': 'Catalyst',
    'XP': 'Catsxp',
    '0C': 'Cave Browser',
    'CL': 'CCleaner',
    'C8': 'CG Browser',
    'CJ': 'ChanjetCloud',
    'C6': 'Chedot',
    'C9': 'Cherry Browser',
    'C0': 'Centaury',
    'ZC': 'Clario Browser',
    '9C': 'Clear TV Browser',
    'CQ': 'Cliqz',
    'CC': 'Coc Coc',
    'C4': 'CoolBrowser',
    'C2': 'Colibri',
    '6C': 'Columbus Browser',
    'CD': 'Comodo Dragon',
    'XC': 'Comet',
    'C1': 'Coast',
    'CX': 'Charon',
    'G0': 'ChatGPT Atlas',
    'CE': 'CM Browser',
    'C7': 'CM Mini',
    'CF': 'Chrome Frame',
    'HC': 'Headless Chrome',
    'CH': 'Chrome',
    'CI': 'Chrome Mobile iOS',
    'CK': 'Conkeror',
    'CM': 'Chrome Mobile',
    '3C': 'Chowbo',
    '7C': 'Classilla',
    'CN': 'CoolNovo',
    '4C': 'Colom Browser',
    'CO': 'CometBird',
    '2C': 'Comfort Browser',
    'CB': 'COS Browser',
    'CW': 'Cornowser',
    'C3': 'Chim Lac',
    'CP': 'ChromePlus',
    'CR': 'Chromium',
    'C5': 'Chromium GOST',
    'CY': 'Cyberfox',
    'CS': 'Cheshire',
    '8C': 'Cromite',
    'RC': 'Crow Browser',
    'CT': 'Crusta',
    'CG': 'Craving Explorer',
    'CZ': 'Crazy Browser',
    'CU': 'Cunaguaro',
    'CV': 'Chrome Webview',
    'YC': 'CyBrowser',
    'DB': 'dbrowser',
    'PD': 'Peeps dBrowser',
    'DK': 'Dark Web',
    'DP': 'Dark Web Private',
    'D4': 'DDU Browser',
    'D1': 'Debuggable Browser',
    'DC': 'Decentr',
    'DE': 'Deepnet Explorer',
    'DG': 'deg-degan',
    'DA': 'Deledao',
    'DT': 'Delta Browser',
    'D0': 'Desi Browser',
    'DS': 'DeskBrowse',
    'D3': 'Dezor',
    'II': 'Diigo Browser',
    'D2': 'DoCoMo',
    'DF': 'Dolphin',
    'DZ': 'Dolphin Zero',
    'DO': 'Dorado',
    'DR': 'Dot Browser',
    'DL': 'Dooble',
    'D8': 'Doubao',
    'DI': 'Dillo',
    'DU': 'DUC Browser',
    'DD': 'DuckDuckGo Privacy Browser',
    'E1': 'East Browser',
    'EC': 'Ecosia',
    'EW': 'Edge WebView',
    'EV': 'Every Browser',
    'EI': 'Epic',
    'EL': 'Elinks',
    'EN': 'EinkBro',
    'EB': 'Element Browser',
    'EE': 'Elements Browser',
    'EO': 'Eolie',
    'EX': 'Explore Browser',
    'EZ': 'eZ Browser',
    'E2': 'EudoraWeb',
    'EU': 'EUI Browser',
    'EP': 'GNOME Web',
    'G1': 'G Browser',
    'ES': 'Espial TV Browser',
    'FG': 'fGet',
    'FA': 'Falkon',
    'FX': 'Faux Browser',
    'F8': 'Fire Browser',
    'F4': 'Fiery Browser',
    'F1': 'Firefox Mobile iOS',
    'FB': 'Firebird',
    'FD': 'Fluid',
    'FE': 'Fennec',
    'FF': 'Firefox',
    'FK': 'Firefox Focus',
    'FY': 'Firefox Reality',
    'FR': 'Firefox Rocket',
    '1F': 'Firefox Klar',
    'F0': 'Float Browser',
    'FL': 'Flock',
    'FP': 'Floorp',
    'FO': 'Flow',
    'F2': 'Flow Browser',
    'FM': 'Firefox Mobile',
    '3F': 'FireSend Browser',
    'FW': 'Fireweb',
    'FN': 'Fireweb Navigator',
    'FH': 'Flash Browser',
    'FS': 'Flast',
    'F5': 'Flyperlink',
    'F9': 'FOSS Browser',
    'FU': 'FreeU',
    'F6': 'Freedom Browser',
    'FT': 'Frost',
    'F3': 'Frost+',
    'FI': 'Fulldive',
    'GA': 'Galeon',
    'G8': 'Gener8',
    'GH': 'Ghostery Privacy Browser',
    'GI': 'GinxDroid Browser',
    'GB': 'Glass Browser',
    'GD': 'Godzilla Browser',
    'G3': 'Good Browser',
    'GE': 'Google Earth',
    'GP': 'Google Earth Pro',
    'GO': 'GOG Galaxy',
    'GR': 'GoBrowser',
    'GK': 'GoKu',
    'G2': 'GO Browser',
    'RN': 'GreenBrowser',
    'HW': 'Habit Browser',
    'H7': 'Halo Browser',
    'HB': 'Harman Browser',
    'HS': 'HasBrowser',
    'HA': 'Hawk Turbo Browser',
    'HQ': 'Hawk Quick Browser',
    'HE': 'Helio',
    'HN': 'Herond Browser',
    'HX': 'Hexa Web Browser',
    'H8': 'HeyTapBrowser',
    'HI': 'Hi Browser',
    'HO': 'hola! Browser',
    'H4': 'Holla Web Browser',
    'H5': 'HotBrowser',
    'HJ': 'HotJava',
    'H6': 'HONOR Browser',
    'HT': 'HTC Browser',
    'HU': 'Huawei Browser Mobile',
    'HP': 'Huawei Browser',
    'HM': 'Harmony 360 Browser',
    'H3': 'HUB Browser',
    'IO': 'iBrowser',
    'IS': 'iBrowser Mini',
    'IB': 'IBrowse',
    'I6': 'iDesktop PC Browser',
    'IC': 'iCab',
    'I2': 'iCab Mobile',
    '0I': 'Ifbrowser',
    '4I': 'iNet Browser',
    'I1': 'Iridium',
    'I3': 'Iron Mobile',
    'I4': 'IceCat',
    'ID': 'IceDragon',
    'IV': 'Isivioo',
    'I8': 'IVVI Browser',
    'IW': 'Iceweasel',
    '2I': 'Impervious Browser',
    'N3': 'Incognito Browser',
    'IN': 'Inspect Browser',
    'I9': 'Insta Browser',
    'IE': 'Internet Explorer',
    'I7': 'Internet Browser Secure',
    '5I': 'Internet Webbrowser',
    '3I': 'Intune Managed Browser',
    'I5': 'Indian UC Mini Browser',
    'Z0': 'InBrowser',
    'IG': 'Involta Go',
    'IM': 'IE Mobile',
    'IR': 'Iron',
    'JB': 'Japan Browser',
    'JS': 'Jasmine',
    'JA': 'JavaFX',
    'JL': 'Jelly',
    'JI': 'Jig Browser',
    'JP': 'Jig Browser Plus',
    'JO': 'JioSphere',
    'JZ': 'JUZI Browser',
    'KB': 'K.Browser',
    'KF': 'Keepsafe Browser',
    'K7': 'KeepSolid Browser',
    'KS': 'Kids Safe Browser',
    'KI': 'Kindle Browser',
    'KM': 'K-meleon',
    'KJ': 'K-Ninja',
    'KO': 'Konqueror',
    'KP': 'Kapiko',
    'KE': 'Keyboard Browser',
    'KN': 'Kinza',
    'K4': 'Kitt',
    'KW': 'Kiwi',
    'KD': 'Kode Browser',
    'KU': 'KUN',
    'KT': 'KUTO Mini Browser',
    'KY': 'Kylo',
    'KZ': 'Kazehakase',
    'LB': 'Cheetah Browser',
    'LD': 'Ladybird',
    'LA': 'Lagatos Browser',
    'GN': 'Legan Browser',
    'LM': 'Lemur Browser',
    'LR': 'Lexi Browser',
    'LV': 'Lenovo Browser',
    'LF': 'LieBaoFast',
    'LG': 'LG Browser',
    'LH': 'Light',
    'L4': 'Lightning Browser Plus',
    'L1': 'Lilo',
    'LI': 'Links',
    'RI': 'Liri Browser',
    'LC': 'LogicUI TV Browser',
    'IF': 'Lolifox',
    'L3': 'Lotus',
    'LO': 'Lovense Browser',
    'LT': 'LT Browser',
    'LU': 'LuaKit',
    'LJ': 'LUJO TV Browser',
    'LL': 'Lulumi',
    'LS': 'Lunascape',
    'LN': 'Lunascape Lite',
    'LX': 'Lynx',
    'L2': 'Lynket Browser',
    'MD': 'Mandarin',
    'MP': 'Maple',
    'M5': 'MarsLab Web Browser',
    'M7': 'MaxBrowser',
    'M1': 'mCent',
    'MB': 'MicroB',
    'MC': 'NCSA Mosaic',
    'MZ': 'Meizu Browser',
    'ME': 'Mercury',
    'M2': 'Me Browser',
    'MF': 'Mobile Safari',
    'MI': 'Midori',
    'M3': 'Midori Lite',
    'M6': 'MixerBox AI',
    'MO': 'Mobicip',
    'MU': 'Mi Browser',
    'MS': 'Mobile Silk',
    'MK': 'Mogok Browser',
    'M8': 'Motorola Internet Browser',
    'MN': 'Minimo',
    'MT': 'Mint Browser',
    'MX': 'Maxthon',
    'M4': 'MaxTube Browser',
    'MA': 'Maelstrom',
    '3M': 'Mises',
    'MM': 'Mmx Browser',
    'NM': 'MxNitro',
    'MY': 'Mypal',
    'MR': 'Monument Browser',
    'MW': 'MAUI WAP Browser',
    'N7': 'Naenara Browser',
    'NW': 'Navigateur Web',
    'NK': 'Naked Browser',
    'NA': 'Naked Browser Pro',
    '2N': 'Neuro Browser',
    'NR': 'NFS Browser',
    'N5': 'Ninetails',
    'NB': 'Nokia Browser',
    'NO': 'Nokia OSS Browser',
    'NV': 'Nokia Ovi Browser',
    'N2': 'Norton Private Browser',
    'NX': 'Nox Browser',
    'N1': 'NOMone VR Browser',
    'N6': 'Nook Browser',
    'NE': 'NetSurf',
    'NF': 'NetFront',
    'NL': 'NetFront Life',
    'NP': 'NetPositive',
    'NS': 'Netscape',
    'WR': 'NextWord Browser',
    'N8': 'Ninesky',
    '3N': 'Nintendo Browser',
    'NT': 'NTENT Browser',
    'NU': 'Nuanti Meta',
    'NI': 'Nuviu',
    'O9': 'Ocean Browser',
    'OC': 'Oculus Browser',
    'O6': 'Odd Browser',
    'O1': 'Opera Mini iOS',
    'OB': 'Obigo',
    'O2': 'Odin',
    '2O': 'Odin Browser',
    'H2': 'OceanHero',
    'OD': 'Odyssey Web Browser',
    'OF': 'Off By One',
    'O5': 'Office Browser',
    'HH': 'OhHai Browser',
    'OL': 'OnBrowser Lite',
    'OE': 'ONE Browser',
    'N4': 'Onion Browser',
    '1N': 'ONIONBrowser',
    '6O': 'Opera Air',
    'Y1': 'Opera Crypto',
    'OX': 'Opera GX',
    'OG': 'Opera Neon',
    'OH': 'Opera Devices',
    'OI': 'Opera Mini',
    'OM': 'Opera Mobile',
    'OP': 'Opera',
    'ON': 'Opera Next',
    'OO': 'Opera Touch',
    'OU': 'Orbitum',
    'OA': 'Orca',
    'OS': 'Ordissimo',
    'OR': 'Oregano',
    'O0': 'Origin In-Game Overlay',
    'OY': 'Origyn Web Browser',
    'O8': 'OrNET Browser',
    'OV': 'Openwave Mobile Browser',
    'O3': 'OpenFin',
    'O4': 'Open Browser',
    '4U': 'Open Browser 4U',
    '5G': 'Open Browser fast 5G',
    '5O': 'Open Browser Lite',
    'O7': 'Open TV Browser',
    'OW': 'OmniWeb',
    'OT': 'Otter Browser',
    '4O': 'Owl Browser',
    'JR': 'OJR Browser',
    'PL': 'Palm Blazer',
    'PM': 'Pale Moon',
    'PY': 'Polypane',
    'PZ': 'Panda Browser',
    '8P': 'Prism',
    'PP': 'Oppo Browser',
    'P6': 'Opus Browser',
    'PR': 'Palm Pre',
    '2E': 'Pocket Internet Explorer',
    '7I': 'Puffin Cloud Browser',
    '6I': 'Puffin Incognito Browser',
    'PU': 'Puffin Secure Browser',
    '2P': 'Puffin Web Browser',
    'PW': 'Palm WebPro',
    'PA': 'Palmscape',
    'P7': 'Pawxy',
    '0P': 'Peach Browser',
    'PE': 'Perfect Browser',
    'K6': 'Perk',
    'P1': 'Phantom.me',
    'PH': 'Phantom Browser',
    'PX': 'Phoenix',
    'PB': 'Phoenix Browser',
    '5P': 'Photon',
    'N9': 'Pintar Browser',
    'P9': 'PirateBrowser',
    'P8': 'PICO Browser',
    'PF': 'PlayFree Browser',
    'PK': 'PocketBook Browser',
    'PO': 'Polaris',
    'PT': 'Polarity',
    'LY': 'PolyBrowser',
    'PQ': 'Power Browser',
    '9P': 'Presearch',
    'BP': 'Privacy Browser',
    'PI': 'PrivacyWall',
    'P4': 'Privacy Explorer Fast Safe',
    'X5': 'Cloak Private Browser',
    'P3': 'Private Internet Browser',
    'P5': 'Proxy Browser',
    '7P': 'Proxyium',
    '6P': 'Proxynet',
    '2F': 'ProxyFox',
    '2M': 'ProxyMax',
    'P2': 'Pi Browser',
    'P0': 'PronHub Browser',
    'PC': 'PSI Secure Browser',
    'RW': 'Reqwireless WebViewer',
    'RO': 'Roccat',
    'PS': 'Microsoft Edge',
    'QA': 'Qazweb',
    'QI': 'Qiyu',
    'QJ': 'QJY TV Browser',
    'Q3': 'Qmamu',
    'Q4': 'Quick Search TV',
    'Q2': 'QQ Browser Lite',
    'Q1': 'QQ Browser Mini',
    'QQ': 'QQ Browser',
    'QS': 'Quick Browser',
    'QT': 'Qutebrowser',
    'QU': 'Quark',
    'Q6': 'QuarkPC',
    'Q7': 'Quetta',
    'QZ': 'QupZilla',
    'QM': 'Qwant Mobile',
    'Q5': 'QtWeb',
    'QW': 'QtWebEngine',
    'R3': 'Rakuten Browser',
    'R4': 'Rakuten Web Search',
    'R2': 'Raspbian Chromium',
    'R6': 'Ray',
    'RT': 'RCA Tor Explorer',
    'RE': 'Realme Browser',
    'RK': 'Rekonq',
    'RM': 'RockMelt',
    'RB': 'Roku Browser',
    'SB': 'Samsung Browser',
    '3L': 'Samsung Browser Lite',
    'SA': 'Sailfish Browser',
    'R0': 'SberBrowser',
    'S8': 'Seewo Browser',
    'SC': 'SEMC-Browser',
    'SE': 'Sogou Explorer',
    'SO': 'Sogou Mobile Browser',
    'RF': 'SOTI Surf',
    '2S': 'Soul Browser',
    'T0': 'Soundy Browser',
    'SF': 'Safari',
    'PV': 'Safari Technology Preview',
    'S5': 'Safe Exam Browser',
    'SW': 'SalamWeb',
    'VN': 'Savannah Browser',
    'SD': 'SavySoda',
    'S9': 'Secure Browser',
    'SV': 'SFive',
    'SH': 'Shiira',
    'K1': 'Sidekick',
    'S1': 'SimpleBrowser',
    '3S': 'SilverMob US',
    'ZB': 'Singlebox',
    'SY': 'Sizzy',
    'K3': 'Skye',
    'SK': 'Skyfire',
    'KL': 'SkyLeap',
    'SJ': 'Seekee',
    'SS': 'Seraphic Sraf',
    'KK': 'SiteKiosk',
    'SL': 'Sleipnir',
    '8B': 'SlimBoat',
    'S6': 'Slimjet',
    'S7': 'SP Browser',
    '9S': 'Sony Small Browser',
    '8S': 'Secure Private Browser',
    'X2': 'SecureX',
    'T1': 'Stampy Browser',
    '7S': '7Star',
    'SQ': 'Smart Browser',
    '6S': 'Smart Search & Web Browser',
    'LE': 'Smart Lenovo Browser',
    'OZ': 'Smooz',
    'SM': 'Snap Browser',
    'SN': 'Snowshoe',
    'K5': 'Spark',
    'B1': 'Spectre Browser',
    'S2': 'Splash',
    'SI': 'Sputnik Browser',
    'SR': 'Sunrise',
    '0S': 'Sunflower Browser',
    'SP': 'SuperBird',
    'SU': 'Super Fast Browser',
    '5S': 'SuperFast Browser',
    'HR': 'Sushi Browser',
    'S3': 'surf',
    '4S': 'Surf Browser',
    'RY': 'Surfy Browser',
    'SG': 'Stargon',
    'S0': 'START Internet Browser',
    'YS': 'Stay Browser',
    '5A': 'Stealth Browser',
    'S4': 'Steam In-Game Overlay',
    'ST': 'Streamy',
    'SX': 'Swiftfox',
    'W7': 'Swiftweasel',
    'SZ': 'Seznam Browser',
    'W1': 'Sweet Browser',
    '2X': 'SX Browser',
    'TP': 'T+Browser',
    'TR': 'T-Browser',
    'TO': 't-online.de Browser',
    'TT': 'TalkTo',
    'TA': 'Tao Browser',
    'T2': 'tararia',
    'TH': 'Thor',
    '1T': 'Tor Browser',
    'TF': 'TenFourFox',
    'TB': 'Tenta Browser',
    'TE': 'Tesla Browser',
    'TZ': 'Tizen Browser',
    'T5': 'Tincat Browser',
    'TI': 'Tint Browser',
    'TL': 'TrueLocation Browser',
    'TC': 'TUC Mini Browser',
    'TK': 'TUSK',
    'TU': 'Tungsten',
    'TG': 'ToGate',
    'T3': 'Total Browser',
    'TQ': 'TQ Browser',
    'TS': 'TweakStyle',
    'TV': 'TV Bro',
    'T4': 'TV-Browser Internet',
    'U0': 'U Browser',
    'UB': 'UBrowser',
    'UC': 'UC Browser',
    'UH': 'UC Browser HD',
    'UM': 'UC Browser Mini',
    'UT': 'UC Browser Turbo',
    'UI': 'Ui Browser Mini',
    'UP': 'UPhone Browser',
    'UR': 'UR Browser',
    'UZ': 'Uzbl',
    'UE': 'Ume Browser',
    'V0': 'vBrowser',
    'VA': 'Vast Browser',
    'VQ': 'VC Browser Mini Pro',
    'V3': 'VD Browser',
    'VR': 'Veera',
    'VE': 'Venus Browser',
    'WD': 'Vewd Browser',
    'V5': 'VibeMate',
    'NQ': 'Nova Browser',
    'N0': 'Nova Video Downloader Pro',
    'VS': 'Viasat Browser',
    'VI': 'Vivaldi',
    'VV': 'vivo Browser',
    'V2': 'Vivid Browser Mini',
    'VB': 'Vision Mobile Browser',
    'V4': 'Vertex Surf',
    'VM': 'VMware AirWatch',
    'V6': 'VMS Mosaic',
    'VK': 'Vonkeror',
    'VU': 'Vuhuv',
    'WI': 'Wear Internet Browser',
    'WP': 'Web Explorer',
    'W3': 'Web Browser & Explorer',
    'W5': 'Webian Shell',
    'W4': 'WebDiscover',
    'WE': 'WebPositive',
    'W6': 'Weltweitimnetz Browser',
    'WX': 'Wexond',
    'WF': 'Waterfox',
    'WB': 'Wave Browser',
    'WA': 'Wavebox',
    'WH': 'Whale Browser',
    'W2': 'Whale TV Browser',
    'WO': 'wOSBrowser',
    '3W': 'w3m',
    'WT': 'WeTab Browser',
    '1W': 'World Browser',
    'WL': 'Wolvic',
    'WK': 'Wukong Browser',
    'WY': 'Wyzo',
    'YG': 'YAGI',
    'YJ': 'Yahoo! Japan Browser',
    'YA': 'Yandex Browser',
    'Y4': 'Yandex Browser Corp',
    'YL': 'Yandex Browser Lite',
    'YN': 'Yaani Browser',
    'Y2': 'Yo Browser',
    'YB': 'Yolo Browser',
    'YO': 'YouCare',
    'Y3': 'YouBrowser',
    'YZ': 'Yuzu Browser',
    'XR': 'xBrowser',
    'X3': 'MMBOX XBrowser',
    'XB': 'X Browser Lite',
    'X0': 'X-VPN',
    'X1': 'xBrowser Pro Super Fast',
    'XN': 'XNX Browser',
    'X6': 'XPlay Browser',
    'XT': 'XtremeCast',
    'XS': 'xStand',
    'XI': 'Xiino',
    'X4': 'XnBrowse',
    'XO': 'Xooloo Internet',
    'XV': 'Xvast',
    'ZE': 'Zetakey',
    'ZV': 'Zvu',
    'ZI': 'Zirco Browser',
    'ZR': 'Zordo Browser',
    'ZT': 'ZTE Browser',
}

# flip Abbrev / Name for fast membership testing
BROWSER_TO_ABBREV = {browser.lower(): abbrev for abbrev, browser in AVAILABLE_BROWSERS.items()}

# fmt: off
BROWSER_FAMILIES = {
    'Android Browser': ('AN',),
    'BlackBerry Browser': ('BB',),
    'Baidu': ('BD', 'BS', 'H6'),
    'Amiga': ('AV', 'AW'),
    'Chrome': (
        'CH', '2B', '7S', 'A0', 'AC', 'A4', 'AE', 'AH', 'AI',
        'AO', 'AS', 'BA', 'BM', 'BR', 'C2', 'C3', 'C5', 'C4',
        'C6', 'CC', 'CD', 'CE', 'CF', 'CG', '1B', 'CI', 'CL',
        'CM', 'CN', 'CP', 'CR', 'CV', 'CW', 'DA', 'DD', 'DG',
        'DR', 'EC', 'EE', 'EU', 'EW', 'FA', 'FS', 'GB', 'GI',
        'H2', 'HA', 'HE', 'HH', 'HS', 'I3', 'IR', 'JB', 'KN',
        'KW', 'LF', 'LL', 'LO', 'M1', 'MA', 'MD', 'MR', 'MS',
        'MT', 'MZ', 'NM', 'NR', 'O0', 'O2', 'O3', 'OC', 'PB',
        'PT', 'QU', 'QW', 'RM', 'S4', 'S6', 'S8', 'S9', 'SB',
        'SG', 'SS', 'SU', 'SV', 'SW', 'SY', 'SZ', 'T1', 'TA',
        'TB', 'TG', 'TR', 'TS', 'TU', 'TV', 'UB', 'UR', 'VE',
        'VG', 'VI', 'VM', 'WP', 'WH', 'XV', 'YJ', 'FH', 'H8',
        'B1', 'BO', 'HB', 'PC', 'LA', 'LT', 'PD', 'HR', 'HU',
        'HP', 'IO', 'TP', 'CJ', 'HQ', 'HI', 'PN', 'BW', 'YO',
        'DC', 'G8', 'DT', 'AP', 'AK', 'UI', 'SD', 'VN', '4S',
        '2S', 'RF', 'LR', 'SQ', 'BV', 'L1', 'F0', 'KS', 'V0',
        'C8', 'AZ', 'MM', 'BT', 'N0', 'P0', 'F3', 'VS', 'DU',
        'D0', 'P1', 'O4', '8S', 'H3', 'TE', 'WB', 'K1', 'P2',
        'XO', 'U0', 'B0', 'VA', 'X0', 'NX', 'O5', 'R1', 'I1',
        'HO', 'A5', 'X1', '18', 'B5', 'B6', 'TC', 'A6', '2X',
        'F4', 'YG', 'WR', 'NA', 'DM', '1M', 'A7', 'XN', 'XT',
        'XB', 'W1', 'HT', 'B8', 'F5', 'B9', 'WA', 'T0', 'HC',
        'O6', 'P7', 'LJ', 'LC', 'O7', 'N2', 'A8', 'P8', 'RB',
        '1W', 'EV', 'I9', 'V4', 'H4', '1T', 'M5', '0S', '0C',
        'ZR', 'D6', 'F6', 'RC', 'WD', 'P3', 'FT', 'A9', 'X2',
        'N3', 'GD', 'O9', 'Q3', 'F7', 'K2', 'P5', 'H5', 'V3',
        'K3', 'Q4', 'G2', 'R2', 'WX', 'XP', '3I', 'BG', 'R0',
        'JO', 'OL', 'GN', 'W4', 'QI', 'E1', 'RI', '8B', '5B',
        'K4', 'WK', 'T3', 'K5', 'MU', '9P', 'K6', 'VR', 'N9',
        'M9', 'F9', '0P', '0A', 'JR', 'D3', 'TK', 'BP', '2F',
        '2M', 'K7', '1N', '8A', 'H7', 'X3', 'T4', 'X4', '5O',
        '8C', '3M', '6I', '2P', 'PU', '7I', 'X5', 'AL', '3P',
        'W2', 'ZB', 'HN', 'Q6', 'Q7', 'G0', '00', 'R6', 'D8',
        'PQ', 'LM', 'T5', '2N', 'SJ', 'X6', 'SM', 'AY', 'BQ',
        'BC', 'NQ', 'VQ', '9C', 'KA', 'YS', 'D4', 'PZ', '0I',
        '3F', 'Z1', 'XC', 'ZC',
    ),
    'Firefox': (
        'FF', 'BI', 'BF', 'BH', 'BN', 'C0', 'CU', 'EI', 'F1',
        'FB', 'FE', 'AX', 'FM', 'FR', 'FY', 'I4', 'IF', '8P',
        'IW', 'LH', 'LY', 'MB', 'MN', 'MO', 'MY', 'OA', 'OS',
        'PI', 'PX', 'QA', 'S5', 'SX', 'TF', 'TO', 'WF', 'ZV',
        'FP', 'AD', '2I', 'P9', 'KJ', 'WY', 'VK', 'W5',
        '7C', 'N7', 'W7',
    ),
    'Internet Explorer': ('IE', 'CZ', 'BZ', 'IM', 'PS', '3A', '4A', 'RN', '2E'),
    'Konqueror': ('KO',),
    'NetFront': ('NF', '3N'),
    'NetSurf': ('NE',),
    'Nokia Browser': ('DO', 'NB', 'NO', 'NV'),
    'Opera': (
        'OP', 'OG', 'OH', 'OI', 'OM', 'ON', 'OO', 'O1', 'OX',
        'Y1', '6O',
    ),
    'Safari': ('SF', 'S7', 'MF', 'SO', 'PV'),
    'Sailfish Browser': ('SA',),
}

FAMILY_FROM_ABBREV = {}
for b, codes in BROWSER_FAMILIES.items():
    for code in codes:
        FAMILY_FROM_ABBREV[code] = b.lower()

MOBILE_ONLY_BROWSERS = {
    '36', 'AH', 'AI', 'BL', 'C1', 'C4', 'CB', 'CW', 'DB',
    '3M', 'DT', 'EU', 'EZ', 'FK', 'FM', 'FR', 'FX', 'GH',
    'GI', 'GR', 'HA', 'HU', 'IV', 'JB', 'KD', 'M1', 'MF',
    'MN', 'MZ', 'NX', 'OC', 'OI', 'OM', 'OZ', '2P', 'PI',
    'PE', 'QU', 'RE', 'S0', 'S7', 'SA', '0I', 'SG', 'SK',
    'ST', 'SU', 'T1', 'UH', 'UM', 'UT', 'VE', 'VV', 'WI',
    'WP', 'YN', 'IO', 'IS', 'HQ', 'RW', 'HI', 'PN', 'BW',
    'YO', 'PK', 'MR', 'AP', 'AK', 'UI', 'SD', 'VN', '4S',
    'RF', 'LR', 'SQ', 'BV', 'L1', 'F0', 'KS', 'V0', 'C8',
    'AZ', 'MM', 'BT', 'N0', 'P0', 'F3', 'DU', 'D0', 'P1',
    'O4', 'XO', 'U0', 'B0', 'VA', 'X0', 'A5', 'X1', '18',
    'B5', 'B6', 'TC', 'A6', '2X', 'F4', 'YG', 'WR', 'NA',
    'DM', '1M', 'A7', 'XN', 'XT', 'XB', 'W1', 'HT', 'B7',
    'B9', 'T0', 'I8', 'O6', 'P7', 'O8', '4B', 'A8', 'P8',
    '1W', 'EV', 'Z0', 'I9', 'V4', 'H4', 'M5', '0S', '0C',
    'ZR', 'D6', 'F6', 'P3', 'FT', 'A9', 'X2', 'NI', 'FG',
    'TH', 'N3', 'GD', 'O9', 'Q3', 'F7', 'K2', 'N4', 'P5',
    'H5', 'V3', 'G2', 'BG', 'OL', 'II', 'TL', 'M6', 'Y3',
    'M7', 'GN', 'JR', 'IG', 'HW', '4O', 'OU', '5P', 'KE',
    '5A', 'TT', '6P', 'G3', '7P', 'VU', 'F8', 'L4', 'DK',
    'DP', 'KL', 'K4', 'N6', 'KU', 'WK', 'M8', 'UP', 'ZT',
    '9P', 'N8', 'VR', 'N9', 'M9', 'F9', '0P', '0A', '2F',
    '2M', 'K7', '1N', '8A', 'H7', 'X3', 'X4', '5O', '6I',
    '7I', 'X5', '3P', '2E', 'T5', '2N', 'SJ', 'X6', 'SM',
    'AY', 'BQ', 'BC', 'NQ', 'VQ', 'KA', 'YS', 'D4', 'PZ',
    '0I',
}
# fmt: off

TV_CLIENTS = {
    'Kylo', 'Espial TV Browser', 'LUJO TV Browser', 'LogicUI TV Browser', 'Open TV Browser',
    'Seraphic Sraf', 'Opera Devices', 'Crow Browser', 'Vewd Browser', 'TiviMate',
    'Quick Search TV', 'QJY TV Browser', 'TV Bro',
    'Redline',
}

# Taken from ChromeOS regex in oss.yml
CHROME_OS_MODELS = {
    'asurada', 'atlas', 'brask', 'brox', 'brya', 'cherry', 'coral', 'corsola',
    'dedede', 'drallion', 'elm', 'eve', 'fizz', 'geralt', 'grunt', 'guybrush',
    'hana', 'hatch', 'jacuzzi', 'kalista', 'kukui', 'nami', 'nautilus', 'nirva',
    'nissa', 'nocturne', 'octopus', 'ovis', 'puff', 'pyro', 'rammus', 'rauru',
    'reef', 'rex', 'sand', 'sarien', 'skyrim', 'snappy', 'soraka', 'staryu',
    'strongbad', 'trogdor', 'volteer', 'zork',
}

# Crufty name/version prefixes too generic to be meaningful Client names
# iOS/12.1, CFNetwork/975.0.3, Android/7.0
CRUFT_NAMES = {
    'afma-sdk-a',
    '1.0,win',
    'alamofire',
    'applewebkit',
    'carrier',
    'cfnetwork',
    'client',
    'client-protocol',
    'configuration',
    'darwin',
    'dalvik',
    'hw',
    'j2me',
    'mozilla',
    'mozzila',
    'user-agent:mozilla',
    'android mozilla',
    'mobile',
    'ios',
    'android',
    'ipad',
    'iphone',
    'okhttp',
    'oem',
    'os',
    'opera mobi',
    'profile',
    'scale',
    'sdk',
    'sid',
    'sparkle',
    'ss',
    'winsparkle',
    'symbian',
    'unknown',
    'urbanairshiplib',
    'urbanairshiplib-android',
    'windows',
    'windows nt',
    'win64',
    'wow64',
}

# These keys may provide meaningful information,
# but should never be considered app names.
METADATA_NAMES = {
    'app.webview',
    'app.webview app.webview',
    'build',
    'device',
    'ioswebview',
    'linux',
    'version',
    'wfappid',
    'wfexperience',
    'macosx',
    'x-client',
    'windows',
}

# When parsing UA strings generically, multiple name/version pairs may be found.
# Ignore the uninteresting ones
# Mozilla/5.0 (Symbian/3; Series60/5.2 NokiaN8-00/014.002; Profile/MIDP-2.1 Configuration/CLDC-1.1; en-us) AppleWebKit/525 (KHTML, like Gecko) Version/3.0 BrowserNG/7.2.6.4 3gpp-gba  # noqa
SKIP_PREFIXES = set(AVAILABLE_ENGINES_LOWER_CASE.keys()) | \
                CRUFT_NAMES | \
                set(BROWSER_TO_ABBREV.keys())

CHECK_PAIRS = {
    'Android Browser',
    'Mobile Safari',
    'Chrome Mobile',
    'Chrome',
}


def normalized_name(name: str, abbreviations: dict[str, str], names: dict[str, str]) -> str:
    """
    Lookup Brand name by lower casing to account
    for spelling differences.
    """
    try:
        abbrev = abbreviations[str(name).strip().lower()]
        return names[abbrev]
    except KeyError:
        pass
    return name


__all__ = (
    'APPLE_OS_NAMES',
    'APP_TO_CLIENT_HINT_MAPPING',
    'CHROME_OS_MODELS',
    'CLIENT_HINT_TO_APP_MAP',
    'OS_CLIENT_HINT_MAPPING',
    'ALWAYS_DESKTOP_OS',
    'DESKTOP_OS',
    'OPERATING_SYSTEMS',
    'OS_TO_ABBREV',
    'OS_FAMILIES',
    'FAMILY_FROM_OS',
    'AVAILABLE_ENGINES',
    'AVAILABLE_BROWSERS',
    'BROWSER_FAMILIES',
    'BROWSER_TO_ABBREV',
    'FAMILY_FROM_ABBREV',
    'CHECK_PAIRS',
    'CRUFT_NAMES',
    'METADATA_NAMES',
    'MOBILE_ONLY_BROWSERS',
    'TV_CLIENTS',
    'SKIP_PREFIXES',
    'normalized_name',
)



from typing import Any
from . import BaseDeviceParser
from ...lazy_regex import RegexLazyIgnore


class VendorFragment(BaseDeviceParser):
    __slots__ = ()
    fixture_files = [
        'upstream/vendorfragments.yml',
    ]

    def yaml_to_list(self, yfile: str) -> list[dict[str, Any]]:
        """
        List of dicts like so:

        {'brand': 'Dell', 'regexes': ['MDDR(JS)?', 'MDDC(JS)?', 'MDDS(JS)?']}
        """
        new_regexes = self.load_from_yaml(yfile)
        if isinstance(new_regexes, list):
            return new_regexes

        reg_list = []

        for brand, regexes in new_regexes.items():
            reg_list.append({
                'brand': brand,
                'regexes': [RegexLazyIgnore(r) for r in regexes],
            })

        return reg_list

    def _parse(self) -> None:
        user_agent = self.user_agent
        for ua_data in self.regex_list:
            for vendor in ua_data['regexes']:
                if matched := vendor.search(user_agent):
                    self.matched_regex = matched
                    self.ua_data = {k: v for k, v in ua_data.items() if k != 'regexes'}
                    self.known = True

                    return


__all__ = [
    'VendorFragment',
]



from urllib.parse import unquote
from ..base import ParserBaseTest
from ...parser import OS
from ...device_detector import DeviceDetector
from ...parser.client_hints import ClientHints


class TestClientHints(ParserBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/clienthints.yml',
        'tests/fixtures/upstream/clienthints-app.yml',
    ]

    def test_parsing(self):
        for fixture in self.load_fixtures():
            self.user_agent = unquote(fixture['user_agent'])
            parsed = DeviceDetector(user_agent=self.user_agent, headers=fixture['headers']).parse()
            parsed_data = parsed.all_details

            # Skip checking the "device" section since
            # we're not ss interested in device specifics.
            for section in ('os', 'client'):
                expected = fixture.get(section) or {}
                extracted = parsed_data.get(section) or {}

                for k, expected_value in expected.items():
                    if k in ('engine', 'version', 'engine_version'):
                        continue

                    extracted_value = None if extracted.get(k) == 'Unknown' else extracted.get(k)

                    # Consider names valid if they vary only by spacing
                    if k == 'name' and extracted_value:
                        if (
                            extracted_value != expected_value
                            and extracted_value.replace(' ', '') == expected_value.replace(' ', '')
                        ):
                            continue

                    self.assertEqual(
                        extracted_value,
                        extracted_value,
                        msg=f'Section: {section!r}. '
                            f'Key: {k!r}. '
                            f'Expected: {expected_value!r}. '
                            f'Actual: {extracted_value!r}. '
                            f'User agent: {self.user_agent!r}',
                    )

    def test_headers(self):
        headers = {
            'sec-ch-ua' : '"Opera";v="83", " Not;A Brand";v="99", "Chromium";v="98"',
            'sec-ch-ua-mobile' : '?0',
            'sec-ch-ua-platform' : 'Windows',
            'sec-ch-ua-platform-version': '14.0.0',
        }
        ch = ClientHints.new(headers)

        self.assertFalse(ch.is_mobile())
        self.assertTrue(ch.is_desktop())
        self.assertEqual(ch.platform, 'Windows')
        self.assertEqual(ch.platform_version, '14.0.0')

        ch_map = {
            'Opera': '83',
            'Chromium': '98',
        }
        self.assertDictEqual(ch_map, ch.client_hints_map)

    def test_headers_http(self):
        headers = {
            'HTTP_SEC_CH_UA_FULL_VERSION_LIST' : '" Not A;Brand";v="99.0.0.0", "Chromium";v="98.0.4758.82", "Opera";v="98.0.4758.82"',
            'HTTP_SEC_CH_UA': '" Not A;Brand";v="99", "Chromium";v="98", "Opera";v="84"',
            'HTTP_SEC_CH_UA_MOBILE' : '?0',
            'HTTP_SEC_CH_UA_MODEL' : 'DN2103',
            'HTTP_SEC_CH_UA_PLATFORM' : 'Ubuntu',
            'HTTP_SEC_CH_UA_PLATFORM_VERSION': '3.7',
            'HTTP_SEC_CH_UA_FULL_VERSION': '98.0.14335.105',
        }
        ch = ClientHints.new(headers)

        self.assertFalse(ch.is_mobile())
        self.assertTrue(ch.is_desktop())
        self.assertEqual(ch.platform, 'Ubuntu')
        self.assertEqual(ch.platform_version, '3.7')
        self.assertEqual(ch.model, 'DN2103')

        ch_map = {
            'Opera': '98.0.4758.82',
            'Chromium': '98.0.4758.82',
        }
        self.assertDictEqual(ch_map, ch.client_hints_map)

    def test_os(self):

        fixtures = [
            {
                'user_agent': 'Mozilla/5.0 (Windows NT 10.0; WOW64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36 OPR/83.0.4254.27',
                'headers': {
                    'sec-ch-ua': '"Opera";v="83", " Not;A Brand";v="99", "Chromium";v="98"',
                    'sec-ch-ua-mobile': '?0',
                    'sec-ch-ua-platform': 'Windows',
                    'sec-ch-ua-platform-version': '14.0.0',
                },
                'os': {
                    'name': 'Windows',
                    'short_name': 'WIN',
                    'version': '11',
                    'family': 'Windows',
                },
            },
            {
                'user_agent': '',
                'headers': {
                    'Sec-CH-UA': '"AvastSecureBrowser";v="6.6.0", " Not A;Brand";v="99.0.0.0", "Chromium";v="98.0.4758.101"',
                    'Sec-CH-UA-Platform': 'Android',
                    'Sec-CH-UA-Mobile': '?1',
                    'Sec-CH-UA-Full-Version': '6.6.0',
                    'Sec-CH-UA-Platform-Version': '11',
                    'Sec-CH-UA-Arch': '[arm64-v8a, armeabi-v7a, armeabi]',
                    'Sec-CH-Prefers-Color-Scheme': 'light',
                },
                'os': {
                    'name': 'Android',
                    'version': '11',
                    'platform': 'ARM',
                },
            },
        ]

        for fixture in fixtures:
            self.user_agent = unquote(fixture.pop('user_agent'))
            expect = fixture['os']

            ch = ClientHints.new(fixture['headers'])
            parsed = OS(self.user_agent, ch).parse()

            data = parsed.ua_data

            for key, value in expect.items():
                self.assertEqual(
                    str(expect[key]),
                    str(data.get(key)),
                    field=key,
                )



from ..base import DetectorBaseTest


class TestDetectSmartPhone(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone.yml',
    ]


class TestDetectSmartPhone1(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-1.yml',
    ]


class TestDetectSmartPhone2(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-2.yml',
    ]


class TestDetectSmartPhone3(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-3.yml',
    ]


class TestDetectSmartPhone4(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-4.yml',
    ]


class TestDetectSmartPhone5(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-5.yml',
    ]


class TestDetectSmartPhone6(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-6.yml',
    ]


class TestDetectSmartPhone7(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-7.yml',
    ]


class TestDetectSmartPhone8(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-8.yml',
    ]


class TestDetectSmartPhone9(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-9.yml',
    ]


class TestDetectSmartPhone10(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-10.yml',
    ]


class TestDetectSmartPhone11(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-11.yml',
    ]


class TestDetectSmartPhone12(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-12.yml',
    ]


class TestDetectSmartPhone13(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-13.yml',
    ]


class TestDetectSmartPhone14(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-14.yml',
    ]


class TestDetectSmartPhone15(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-15.yml',
    ]


class TestDetectSmartPhone16(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-16.yml',
    ]


class TestDetectSmartPhone17(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-17.yml',
    ]


class TestDetectSmartPhone18(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-18.yml',
    ]


class TestDetectSmartPhone19(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-19.yml',
    ]


class TestDetectSmartPhone20(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-20.yml',
    ]


class TestDetectSmartPhone21(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-21.yml',
    ]


class TestDetectSmartPhone22(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-22.yml',
    ]


class TestDetectSmartPhone23(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-23.yml',
    ]


class TestDetectSmartPhone24(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-24.yml',
    ]


class TestDetectSmartPhone25(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-25.yml',
    ]


class TestDetectSmartPhone26(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-26.yml',
    ]


class TestDetectSmartPhone27(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-27.yml',
    ]


class TestDetectSmartPhone28(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-28.yml',
    ]


class TestDetectSmartPhone29(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-29.yml',
    ]


class TestDetectSmartPhone30(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-30.yml',
    ]


class TestDetectSmartPhone31(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-31.yml',
    ]


class TestDetectSmartPhone32(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-32.yml',
    ]


class TestDetectSmartPhone33(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-33.yml',
    ]


class TestDetectSmartPhone34(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-34.yml',
    ]


class TestDetectSmartPhone35(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-35.yml',
    ]


class TestDetectSmartPhone36(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-36.yml',
    ]


class TestDetectSmartPhone37(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-37.yml',
    ]


class TestDetectSmartPhone38(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-38.yml',
    ]


class TestDetectSmartPhone39(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-39.yml',
    ]


class TestDetectSmartPhone40(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-40.yml',
    ]


class TestDetectSmartPhone41(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-41.yml',
    ]


class TestDetectSmartPhone42(DetectorBaseTest):

    fixture_files = [
        'tests/fixtures/upstream/smartphone-42.yml',
    ]



from .base import BaseDeviceParser
from ua_extract.enums import DeviceType


class Camera(BaseDeviceParser):
    __slots__ = ()
    DEVICE_TYPE = DeviceType.Camera

    fixture_files = [
        'upstream/device/cameras.yml',
    ]


__all__ = [
    'Camera',
]



#!/usr/bin/env python3
import io
import os
import re
from setuptools import setup, find_packages


def get_version():
    with open('ua_extract/__init__.py', 'r') as f:
        line = f.readline()
        match = re.match(r'__version__ = \'([\d\.]+)\'', line)

        if not match:
            raise ImportError("Can't read the version of device_detector")

        return match.group(1)


def parse_requirements(filename):
    with open(filename, 'r') as f:
        return [line.strip() for line in f if line.strip() and not line.startswith('#')]


here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

setup(
    name='ua_extract',
    version=get_version(),
    description="Python3 port of matomo's Device Detector",
    long_description=long_description,
    python_requires=">=3.10",
    long_description_content_type='text/markdown',
    author='Pranav Agrawal',
    author_email='pranavagrawal321@gmail.com',
    url='https://github.com/pranavagrawal321/ua_extract',
    packages=find_packages(exclude=["tests"]),
    package_dir={'': '.'},
    license='MIT',
    zip_safe=True,
    keywords=[
        "user agent",
        "app detection",
        "device detection",
        "client hints",
    ],
    include_package_data=True,
    license_files=['LICENSE'],
    package_data={
        '': ['*.yml', '*.sh'],
    },
    install_requires=parse_requirements('requirements.txt'),
    entry_points={
        "console_scripts": [
            "ua_extract=ua_extract.__main__:main",
        ],
    },
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
        'Programming Language :: Python :: 3.14',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    project_urls={
        "Source": "https://github.com/pranavagrawal321/ua_extract",
        "Issues": "https://github.com/pranavagrawal321/ua_extract/issues",
        "Documentation": "https://github.com/pranavagrawal321/ua_extract#readme",
    },
)



__version__ = '1.3.3'
from .settings import *
from .parser import *
from .device_detector import *
from .update_regex import Regexes

import ua_extract.warnings

ua_extract.warnings.enable_colored_warnings()



import regex
from typing import NamedTuple

from ..lazy_regex import RegexLazyIgnore
from .settings import SKIP_PREFIXES

CONTAINS_URL = RegexLazyIgnore(
    r'https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{2,256}\.[a-z]{2,6}\b([-a-zA-Z0-9@:%_\+.~#?&//=]*)',
)

# fmt: off
# Extra version / name from UAs
VERSION_NAME_REGEXES = (
    # 1.172.0.1 - LIVE - Mar 5 2020
    # 17build 113411 LIVE Sep 17 20180
    RegexLazyIgnore(r'(?P<version>[\d\.]+)[ \-]+(?P<name>LIVE)'),

    # 15.5.53 Boxcar
    # 165 CandyCanes
    RegexLazyIgnore(r'^(?P<version>[\d\.]+)[ \-/]+(?P<name>\w+)$'),

    # iPhone/2025.26  iPhone14,5|26.0.1|390|844|3.00|AmericanAirlines
    # iPad/2025.30  iPad8,10|18.6.2|1194|834|2.00|AmericanAirlines
    RegexLazyIgnore(r'\|(?P<version>[\.\d]+).*\|(?P<name>AmericanAirlines)'),

    # Bloomberg|iOS|12.1.2|5.12.7|1d45e9c714c08ddcd8b8d51096f50ee7e2a28060
    # BloombergHorseshoe|iPadOS|15.5|5.77.1|a19956f55bf6a16b94ba5f239226c5ae77f2e3b5
    RegexLazyIgnore(r'(?P<name>[a-zA-Z]+)\|(?:ios|ipados)\|[\.\d]+\|(?P<version>[\d\.\-\w\&\?]+)\|'),

    # EmarsysPredictSDK|osversion:15.3|platform:ios
    RegexLazyIgnore(r'(?P<name>[a-zA-Z]+)\|[a-z\:]+(?P<version>[\d\.]+)'),
)

# Extra name / version from UAs
NAME_VERSION_REGEXES = (

    # Get <key>/<value> or <key> <value> pair from the beginning of the regex
    # generic ua string/1.23.2
    # Call & Chat/1 CFNetwork/902.2 Darwin/17.7.0
    # Microsoft SkyDriveSync 18.091.0506.0007 ship; Windows NT 10.0 (17134)
    # Ensure that User Agents like:
    # Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/Version/25.7.0 Hunt/1398
    # match "Hunt/1398" rather than "25.7.0 Hunt/1398"
    # OneDrive-25.209.1026.0002
    RegexLazyIgnore(r'^(?P<name>[\w\d\.\-_\'!®\?, \+\&]+)[ /-](?P<version>[\d\.]+)'),

    # Extract "Opera Mobi/ADR-25672775" in User Agents like:
    # Opera/9.62 (Android 4.1.2; Linux; Opera Mobi/ADR-25672775) Presto/2.520.13 Version/12.520
    RegexLazyIgnore(r'\b(?P<name>[a-zA-Z\. \-_\'!®\?,\+\&]+)/(?P<version>[\d\.\w-]+)'),

    # Get ALL <key>/<value> pairs from the regex
    RegexLazyIgnore(r'(?P<name>[\w\d\.\-_\'!®\?,\+\&]+)/(?P<version>[\d\.\-\w\&\?]+)\b'),

    # <name><space><version> - anchored at the beginning
    # CarboniteDownloader 6.3.2 build 7466 (Sep-07-2017)
    # libreoffice 5.4.3.2 (92a7159f7e4af62137622921e809f8546db437e5; windows; x86;)
    # openoffice.org 3.2 (320m18(build:9502); windows; x86; bundledlanguages=en-us)
    RegexLazyIgnore(r'^(?P<name>[\w\._\&\'!®\?,\+\&]+) [rv]?(?P<version>[\d\.\-\&\?]+)\b'),

    # <name><sep><version> - anywhere in string
    # Microsoft Office Access 2013 (15.0.4693) Windows NT 6.2, where version == 15.0.4693
    # Microsoft.VisualStudio.Help (2.3)
    # Microsoft URL Control - 6.01.9782
    # openshot-qt-2.4.2
    # DigiCal (v1.8.2b; http://digibites.nl/digical)
    # iPad; 10.3.3; autotrader.com; 3.0.10
    RegexLazyIgnore(
        r'(?P<name>[\w\d\.\-_\&\'!®\?, \+]+) ?[(\-;] ?[vr]?(?P<version>[\d\.]+)(?:\b|\w|$)'
    ),

    # <name><space><version> - anywhere in remainder of string
    # Mozilla/5.0 AppleWebKit/537.36 Mobile Safari/537.36 Android SermonAudio.com 1.9.8, wanting "SermonAudio.com 1.9.8"
    RegexLazyIgnore(r'(?P<name>[\w\._\&\'!®\?,\+\& ]+) [rv]?(?P<version>[\d\.\-\&\?]+)\b'),

    # Get <key><value> pair from beginning of regex, when name & version are not delimited
    # BlueApron2.47.0 (iPhone; iOS 12.1.3; Scale/2.0)
    RegexLazyIgnore(r'^(?P<name>[a-zA-Z\-_\'!®\?, \+\&]+)(?P<version>[\d\.]+)\b'),
)

# <name>/<version> pairs with names matching these
# regexes are not considered interesting
SKIP_NAME_REGEXES = [
    # Android; samsung-SAMSUNG-SM-T377A; 6.0.1; AutoTrader.com; 2.6.4.3.236
    RegexLazyIgnore(r'samsung[- ]sm'),

    # Mozilla/5.0 (iPhone; iPhone103; 12.1.4) MLN/4.30.450041483 (0f3d913a35528d98b8793f4d7aa0539e)"
    RegexLazyIgnore(r'^(iphone|ipad)\d'),
]
# fmt: on


class NameVersion(NamedTuple):
    name: str
    version: str


class CodeNameVersion(NamedTuple):
    code: str
    name: str
    version: str


def name_matches_regex(name: str) -> bool:
    """
    If name matches regex, then don't
    want its name/version pair.
    """
    for rgx in SKIP_NAME_REGEXES:
        if rgx.search(name) is not None:
            return True

    return False


def scrub_name_version_pairs(matches: list[NameVersion]) -> list[CodeNameVersion]:
    """
    Takes list of (name,version) tuples.
    Remove all pairs where name matches SKIP patterns
    """
    pairs = []
    for name, version in matches:
        name = name.strip(' -,')
        if not name:
            continue

        # does this look like base64 encoded data?
        if name.endswith('=='):
            continue

        name_lower = name.lower()
        if name_lower in SKIP_PREFIXES:
            continue

        if name_matches_regex(name):
            continue

        code = name_lower.replace(' ', '')
        pairs.append(CodeNameVersion(code, name, version.strip()))

    return pairs


def extract_version_name_pairs(rgx: regex.Pattern, ua: str) -> list[CodeNameVersion]:
    """
    Extract all key/value pairs of the specified regex,
    where key==version and value==name
    and return pairs along with unmatched portion of ua string.
    """
    if matched := rgx.search(ua):
        name = matched.group('name')
        return scrub_name_version_pairs([NameVersion(name, matched.group('version'))])

    return []


def extract_name_version_pairs(rgx: regex.Pattern, ua: str) -> tuple[list[CodeNameVersion], str]:
    """
    Extract all key/value pairs of the specified regex,
    where key==name and value==version
    and return pairs along with unmatched portion of ua string.
    """
    substring = ua

    if matches := rgx.findall(ua):
        substring = rgx.sub(' ', ua)

    pairs = scrub_name_version_pairs([NameVersion(*pair) for pair in matches])

    return pairs, substring


def key_value_pairs(ua: str) -> list[CodeNameVersion]:
    """
    Extract key/value pairs from User Agent String
    """

    substring = CONTAINS_URL.sub(' ', ua)

    all_pairs = []

    for rgx in VERSION_NAME_REGEXES:
        pairs = extract_version_name_pairs(rgx, substring)
        if pairs:
            all_pairs.extend(pairs)

    # <version>/<name> regexes will be much less common
    # so if we found such entries return then first
    if all_pairs:
        return all_pairs

    for rgx in NAME_VERSION_REGEXES:
        pairs, substring = extract_name_version_pairs(rgx, substring)
        all_pairs.extend(pairs)

    return all_pairs


__all__ = (
    'extract_name_version_pairs',
    'key_value_pairs',
)



import regex
from typing import cast, TypedDict
from ua_extract.enums import DeviceType, AppType
from ..lazy_regex import RegexLazyIgnore
from ..yaml_loader import app_pretty_names_types_data
from .settings import CLIENT_HINT_TO_APP_MAP, FAMILY_FROM_OS, BROWSER_TO_ABBREV


CH_UA = RegexLazyIgnore(r'^"([^"]+)"; ?v="([^"]+)"(?:, )?')
TV_APP = RegexLazyIgnore(r'tv\.?(browser|firefox|search|web)')
CHROMIUM_BASED_BROWSERS = frozenset((
    'Chromium',
    'Chrome Webview',
    'Microsoft Edge',
    'Microsoft Edge Simulate',
))


class ClientHintsHeader(TypedDict):
    platform: str
    platform_version: str
    full_version: str
    full_version_list: str | dict[str, str]
    mobile: str
    architecture: str
    bitness: str
    model: str
    form_factors: set[str]
    app: str


# "Chromium";v="106", "Brave";v="106", "Not;A=Brand";v="99"
# "Not_A Brand";v="8.0.0.0", "Chromium";v="120.0.6099.115", "Google Chrome";v="120.0.6099.115"
# NOT_A_BRAND = {
#     'not;a=brand',
#     'not:a-brand',
#     'not_a brand',
#     'not?a_brand',
#     'not-a.brand',
#     'not(a:brand',
#     'not.a/brand',
#     'not a;brand',
#     ';not a brand',
#     'not=a?brand',
# }

# Normalize forms "Not:A-Brand"
NOT_A_BRAND = regex.compile(r"[^\w\s\']|_| ")
NOT_A_BRAND_FRAGMENT = 'notabrand'


# https://developer.mozilla.org/en-US/docs/Web/HTTP/Reference/Headers/Sec-CH-UA-Form-Factors
MOBILE_FORM_FACTORS = {
    'Mobile',
    'Tablet',
    'EInk',
    'Watch',
}


class ClientHints:
    __slots__ = (
        'platform',
        'platform_version',
        'full_version',
        'full_version_list',
        'mobile',
        'architecture',
        'bitness',
        'model',
        'form_factors',
        'app',
        'headers',
        'client_hints_map',
        '_browser_data',
        '_client_data',
        '_client_name',
        'app_pretty_names',
        '_calculated_app_type',
    )

    def __init__(
        self,
        platform: str,
        platform_version: str,
        full_version: str,
        full_version_list: str | dict[str, str],
        mobile: str,
        architecture: str,
        bitness: str,
        model: str,
        form_factors: set[str],
        app: str,
    ) -> None:
        # fmt: off
        self.platform = platform                     # Sec-CH-UA-Platform
        self.platform_version = platform_version     # Sec-CH-UA-Platform-Version
        self.full_version = full_version             # Sec-CH-UA-Full-Version
        self.full_version_list = full_version_list   # Sec-CH-UA-Full-Version-List
        self.mobile = mobile                         # Sec-CH-UA-Mobile
        self.architecture = architecture             # Sec-CH-UA-Arch
        self.bitness = bitness                       # Sec-CH-UA-Bitness
        self.model = model                           # Sec-CH-UA-Model
        self.form_factors = form_factors             # Sec-CH-UA-Form-Factors
        self.app = app                               # X-Requested-With
        # fmt: on

        self.headers: dict[str, str] = {}
        self.client_hints_map: dict[str, str] = {}

        self._browser_data: dict[str, str] | None = None
        self._client_data: dict[str, str] | None = None
        self._client_name = ''
        self.app_pretty_names = app_pretty_names_types_data()

        # Set from pretty name fixtures, or other Header / UA attributes
        # If this value isn't set, then
        self._calculated_app_type = AppType.Unknown

    def __str__(self) -> str:
        return (
            f'Client: {self.client_name()} '
            f'Platform: {self.platform} ({self.platform_version}) '
            f'Architecture: {self.architecture} ({self.bitness})'
        )

    @classmethod
    def new(cls, headers: dict[str, str]) -> 'ClientHints':
        """
        Generate class from HTTP headers.
        """
        params = ClientHintsHeader(
            platform='',
            platform_version='',
            full_version='',
            full_version_list='',
            mobile='',
            architecture='',
            bitness='',
            model='',
            form_factors=set(),
            app='',
        )

        # fmt: off
        for header_key, value in headers.items():
            key = header_key.lower().replace('_', '-')
            value = value or ''
            match key:
                case (
                    'sec-ch-ua-arch'
                    | 'http-sec-ch-ua-arch'
                    | 'architecture'
                ):
                    params['architecture'] = value.strip('"')

                case (
                    'sec-ch-ua-bitness'
                    | 'http-sec-ch-ua-bitness'
                    | 'bitness'
                ):
                    params['bitness'] = value.strip('"')

                case (
                    'sec-ch-ua-mobile'
                    | 'http-sec-ch-ua-mobile'
                    | 'mobile'
                ):
                    params['mobile'] = value.strip('"')

                case (
                    'sec-ch-ua-model'
                    | 'http-sec-ch-ua-model'
                    | 'model'
                ):
                    params['model'] = value.strip('"')

                # "Brave";v="139"
                case (
                    'sec-ch-ua-full-version'
                    | 'http-sec-ch-ua-full-version'
                    | 'uafullversion'
                    | 'full_version'
                ):
                    params['full_version'] = value.strip('"')

                case (
                    'sec-ch-ua-platform'
                    | 'http-sec-ch-ua-platform'
                    | 'platform'
                ):
                    platform = value.strip('"')
                    params['platform'] = FAMILY_FROM_OS.get(platform, platform)

                case (
                    'sec-ch-ua-platform-version'
                    | 'http-sec-ch-ua-platform-version'
                    | 'platformversion'
                    | 'full_version'
                ):
                    params['platform_version'] = value.strip('"')

                # Not A;Brand";v="99", "Chromium";v="98", "Google Chrome";v="98"
                case (
                    'sec-ch-ua-full-version-list'
                    | 'http-sec-ch-ua-full-version-list'
                    | 'full_version_list'
                ):
                    params['full_version_list'] = value

                case 'brands':
                    if not params['full_version_list'] and isinstance(value, list):
                        params['full_version_list'] = from_ch_list(value)

                # use this only if no other header already set the list
                case 'fullversionlist':
                    if not params['full_version_list'] or 'brands' in headers:
                        if isinstance(value, list):
                            params['full_version_list'] = from_ch_list(value)

                case (
                    'sec-ch-ua'
                    | 'http-sec-ch-ua'
                ):
                    # Don't overwrite full version details truncated version!
                    # " Not A;Brand";v="99.0.0.0", "Chromium";v="98.0.4758.82", "Opera";v="98.0.4758.82"
                    # Not A;Brand";v="99", "Chromium";v="98", "Google Chrome";v="98"
                    if not params['full_version_list']:
                        params['full_version_list'] = value.strip()

                case (
                    'x-requested-with'
                    | 'http-x-requested-with'
                ):
                    if not value.lower().startswith(('xmlhttprequest', 'fetch')):
                        params['app'] = value.strip('"')

                case (
                    'http-sec-ch-ua-form-factors'
                    | 'sec-ch-ua-form-factors' | 'form_factors'
                ):
                    if isinstance(value, (list, set, tuple)):
                        params['form_factors'] = set(value)
                    else:
                        params['form_factors'] = {ff.strip(' "') for ff in value.split(',')}
        # fmt: on

        ch = ClientHints(**params)
        ch.headers = headers
        ch.client_hints_map = from_ch_ua(ch.full_version_list)

        return ch

    def client_name(self) -> str:
        """
        Extract most likely client name from the UA.

        "Not;A=Brand";v="99", "Brave";v="139", "Chromium";v="139"
        """
        if not self._client_name:
            if client_name := self._get_pretty_name(self.app):
                return client_name

            name = extract_name_from_hints(self.client_hints_map)

            # If the name has a "pretty name" override defined
            # use that name & type.
            if not self._calculated_app_type:
                if pretty_name := self._get_pretty_name(name.lower().replace(' ', '')):
                    return pretty_name

            name_lower = name.lower()
            if name_lower in BROWSER_TO_ABBREV:
                self._calculated_app_type = AppType.Browser
            else:
                # Strip "Browser" suffix from browsers such as "Brave Browser"
                # and "CCleaner Browser" per their official name
                if name.endswith(' Browser'):
                    if name_lower[:-8] in BROWSER_TO_ABBREV:
                        name = name[:-8]
                        self._calculated_app_type = AppType.Browser
                elif f'{name_lower} browser' in BROWSER_TO_ABBREV:
                    name = f'{name} Browser'
                    self._calculated_app_type = AppType.Browser

            self._client_name = name

        return self._client_name

    def _get_pretty_name(self, app_name: str) -> str:
        """
        If a fixture defines a "pretty name" for this brand name,
        then set that name and app type.
        """
        if app_name and (app_pretty_name := self.app_pretty_names.get(app_name)):
            self._client_name = app_pretty_name['name']
            self._calculated_app_type = cast(AppType, app_pretty_name['type'])
            return self._client_name
        return ''

    def client_version(self) -> str:
        return self.client_hints_map.get(self.client_name(), '')

    def client_is_browser(self) -> bool:
        return self.client_data().get('type') == AppType.Browser

    def is_mobile(self) -> bool:
        """
        Sec-CH-UA-Mobile
        https://developer.mozilla.org/en-US/docs/Web/HTTP/Reference/Headers/Sec-CH-UA-Mobile
        """
        if self.mobile == '?1':
            return True

        if not self.form_factors:
            return False

        return MOBILE_FORM_FACTORS.issubset(self.form_factors)

    def is_desktop(self) -> bool:
        return self.mobile == '?0' or 'Desktop' in self.client_hints_map

    def is_running_android(self) -> bool:
        return self.mobile == 'Android'

    def is_television(self) -> bool:
        if not self.app:
            return False
        return TV_APP.search(self.app) is not None

    def app_type(self) -> AppType:
        return self._calculated_app_type

    def device_type(self) -> DeviceType | None:
        if self.is_television():
            return DeviceType.TV
        elif self.is_desktop():
            return DeviceType.Desktop
        elif self.is_mobile():
            if 'Tablet' in self.form_factors or 'EInk' in self.form_factors:
                return DeviceType.Tablet
            return DeviceType.Smartphone
        return None

    def client_data(self) -> dict[str, str]:
        """
        Get dictionary of all client data that can
        be extracted from client hints headers.
        """
        if self._client_data is None:
            ch_data = {}

            if name := self.client_name():
                ch_data['name'] = name
            if app_id := self.app:
                ch_data['app_id'] = app_id

            if name or app_id:
                if app_type := self.app_type():
                    ch_data['type'] = app_type
                if version := self.client_version():
                    ch_data['version'] = version
                if abbrev := BROWSER_TO_ABBREV.get(name.lower()):
                    ch_data['short_name'] = abbrev

            self._client_data = ch_data

        return self._client_data or {}

    def os_data(self) -> dict[str, str]:
        """
        Get dictionary of all OS data that can
        be extracted from client hints headers.
        """
        ch_data = {}

        if platform := self.platform:
            ch_data['platform'] = platform
        if platform_version := self.platform_version:
            ch_data['platform_version'] = platform_version

        return ch_data


def from_ch_ua(ua: str | dict[str, str]) -> dict[str, str]:
    """
    Extract values from Client Hint User Agent.

    "Not;A=Brand";v="99", "Brave";v="139", "Chromium";v="139"
    """
    if isinstance(ua, dict):
        return ua

    ch_map = {}
    for segment in ua.split(', '):
        if ch := CH_UA.findall(segment):
            name = ch[0][0]
            if NOT_A_BRAND.sub('', name.lower().strip()) == NOT_A_BRAND_FRAGMENT:
                continue
            dd_name = CLIENT_HINT_TO_APP_MAP.get(name, name)
            ch_map[dd_name] = ch[0][1]

    return ch_map


def from_ch_list(ch: list[dict[str, str]]) -> dict[str, str]:
    """
    Extract values from Client Hints when it's list of dicts.

    [
        {'brand': 'Not_A Brand', 'version': '8'},
        {'brand': 'Chromium', 'version': '120'},
    ]
    """
    ch_map = {}

    for header in ch:
        brand = header['brand']
        if NOT_A_BRAND.sub('', brand.lower().strip()) == NOT_A_BRAND_FRAGMENT:
            continue
        dd_name = CLIENT_HINT_TO_APP_MAP.get(brand, brand)
        ch_map[dd_name] = header['version']

    return ch_map


def extract_name_from_hints(hints: dict[str, str]) -> str:
    """
    Extract most specific name from client hints.
    """

    for client_name in hints:
        # If we detected a brand, that is not Chromium, we will use it,
        # otherwise we will look further
        if client_name not in CHROMIUM_BASED_BROWSERS:
            return client_name

    name = ''
    for client_name in hints:
        name = client_name
        if client_name != 'Chromium':
            break

    return name



from .client import *
from .client_hints import ClientHints
from .extractors import *
from .device import *
from .key_value_pairs import *
from .parser import *
from .operating_system import *
from .os_fragment import *
from .settings import *



from typing import Optional
from ua_extract.enums import AppType

try:
    import rapidjson as json
except ImportError:
    import json

from . import BaseClientParser
from ...utils import calculate_dtype


class DictUA(BaseClientParser):
    """
    Some UA strings can be loaded as dicts directly, or with a little parsing.

    {"ac":"CCDesktop_app","av":"4.8.1.435"}
    {"locale":"en_US","appVersion":"5.2.5","os":"iOS","deviceName":"iPhone","deviceType":"phone","iOSApiLevel":"12.1.4","gzip":true,"appId":"KHAiOS","buildNumber":"31667"}
    target=LetGo; appVersion=1.58.0; bundle=com.letgo.ios; build=524; os=iOS 9.2.1; device=Apple iPad4,2; httpLibrary=Alamofire/4.7.3
    """

    __slots__ = ()

    def load_via_json(self) -> dict[str, str]:
        try:
            # sanity check - really shouldn't need to cast to dict.
            # at least 1 UA doesn't crash json.loads but remains a string.
            return dict(json.loads(self.user_agent))
        except Exception:
            return {}

    def parse_key_value_pairs(self) -> dict[str, str]:
        # AppName=iOSProApp;AppId=3;Platform=iOS;Model=iPad Pro 9.7-inch (Wi-Fi Cellular);OSVersion=12.0;Carrier=iPad;AppVersion=3.27.0.4
        try:
            return dict(item.strip().split("=") for item in self.user_agent.split(";"))
        except Exception:
            return {}

    def ua_as_dict(self) -> Optional[dict[str, str]]:
        valid_json = self.load_via_json()
        if valid_json:
            return valid_json

        return self.parse_key_value_pairs()

    def _parse(self) -> None:
        """
        Each subclass may have `appdetails/<name>.yml` file(s) defined
        containing manually specified details for the regex.

        These files before checking regexes, for best performance.
        """
        ua_as_dict = self.ua_as_dict()
        if not ua_as_dict:
            return

        # descending order of interest!
        for name_key in (
            'app',
            'AppName',
            'target',
            'bundle',
            'bundleId',
            'bundleid',
            'BundleID',
            'ac',
            'appId',
        ):
            name = ua_as_dict.get(name_key, '')
            if name:
                self.app_name = name
                self.ua_data['name'] = name
                break

        # don't try to extract a version if we couldn't find a name
        if not self.ua_data:
            return

        # descending order of interest!
        for version_key in (
            'AppVersion',
            'appVersion',
            'version',
            'Version',
            'ver',
            'Ver',
            'av',
        ):
            if version := ua_as_dict.get(version_key, ''):
                self.app_version = version_key
                self.ua_data['version'] = version
                break

    def dtype(self) -> AppType | str:
        return calculate_dtype(app_name=self.app_name) or super().dtype()


__all__ = [
    'DictUA',
]



from . import BaseClientParser
from ua_extract.enums import AppType


class Library(BaseClientParser):
    __slots__ = ()
    APP_TYPE = AppType.Library

    fixture_files = [
        'local/client/libraries.yml',
        'upstream/client/libraries.yml',
    ]


__all__ = [
    'Library',
]



import stat
import shutil
import asyncio
import aiohttp
import tempfile
import subprocess
from enum import Enum
from urllib.parse import urlparse
from pathlib import Path
from datetime import datetime, timezone
from typing import Optional, Callable, Dict, List
from tenacity import retry, stop_after_attempt, wait_exponential, retry_if_exception_type

ROOT_PATH = Path(__file__).resolve().parent


class UpdateMethod(Enum):
    GIT = "git"
    API = "api"


_method_registry: Dict[UpdateMethod, Callable] = {}


def register(method: UpdateMethod):
    def decorator(func):
        _method_registry[method] = func
        return func

    return decorator


def clean_and_format_regexes():
    script_path = (ROOT_PATH / "configure_regex_interpolators.sh").resolve()

    if not script_path.exists():
        raise FileNotFoundError(f"{script_path} does not exist")

    mode = script_path.stat().st_mode
    if not (mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)):
        script_path.chmod(mode | stat.S_IXUSR)

    subprocess.run(
        [str(script_path)],
        check=True,
    )


class Regexes:
    def __init__(
        self,
        upstream_path: str = ROOT_PATH / "regexes" / "upstream",
        repo_url: str = "https://github.com/matomo-org/device-detector.git",
        branch: str = "master",
        sparse_dir: str = "regexes",
        sparse_fixtures_dir: str = "Tests/fixtures",
        fixtures_upstream_path: str = ROOT_PATH / "tests" / "fixtures" / "upstream",
        sparse_client_dir: str = "Tests/Parser/Client/fixtures",
        client_upstream_dir: str = ROOT_PATH
        / "tests"
        / "parser"
        / "fixtures"
        / "upstream"
        / "client",
        sparse_device_dir: str = "Tests/Parser/Device/fixtures",
        device_upstream_dir: str = ROOT_PATH
        / "tests"
        / "parser"
        / "fixtures"
        / "upstream"
        / "device",
        github_token: Optional[str] = None,
        message_callback: Optional[Callable[[str], None]] = None,
    ):
        self.upstream_path = self._validate_path(upstream_path)
        self.repo_url = repo_url
        self.branch = branch
        self.sparse_dir = sparse_dir
        self.sparse_fixtures_dir = sparse_fixtures_dir
        self.fixtures_upstream_path = self._validate_path(fixtures_upstream_path)
        self.sparse_client_dir = sparse_client_dir
        self.client_upstream_dir = self._validate_path(client_upstream_dir)
        self.sparse_device_dir = sparse_device_dir
        self.device_upstream_dir = self._validate_path(device_upstream_dir)
        self.github_token = github_token
        self.message_callback = message_callback or (lambda _: None)

    def _notify(self, message: str):
        self.message_callback(message)

    def _validate_path(self, path) -> Path:
        path = Path(path).resolve()
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _backup_directory(self, path: Path):
        if path.exists():
            backup = path.with_suffix(path.suffix + ".backup")
            shutil.copytree(path, backup, dirs_exist_ok=True)

    def _backup_all_targets(self):
        for p in (
            self.upstream_path,
            self.fixtures_upstream_path,
            self.client_upstream_dir,
            self.device_upstream_dir,
        ):
            self._backup_directory(p)

    def _restore_directory(self, path: Path):
        backup = path.with_suffix(path.suffix + ".backup")
        if not backup.exists():
            self._notify(f"No backup found for {path}, skipping")
            return

        if path.exists():
            shutil.rmtree(path)

        shutil.copytree(backup, path)

    def rollback_regexes(self):
        self._notify("Rolling back regexes...")
        for p in (
            self.upstream_path,
            self.fixtures_upstream_path,
            self.client_upstream_dir,
            self.device_upstream_dir,
        ):
            self._restore_directory(p)
        self._notify("Rollback complete")

    def _clean_dir(self, path: Path):
        if path.exists():
            shutil.rmtree(path)
        path.mkdir(parents=True, exist_ok=True)

    def update_regexes(self, method: str = "git", dry_run: bool = False):
        if dry_run:
            self._notify(f"[Dry Run] Would update regexes using {method}")
            return

        try:
            method_enum = UpdateMethod(method.lower())
        except ValueError:
            raise ValueError(f"Invalid method: {method}")

        updater = _method_registry.get(method_enum)
        if not updater:
            raise RuntimeError(f"No updater registered for {method_enum}")

        self._backup_all_targets()

        try:
            updater(self)
        except Exception:
            self._notify("Update failed — restoring backups")
            self.rollback_regexes()
            raise


@register(UpdateMethod.GIT)
def _update_with_git(self: Regexes):
    if shutil.which("git") is None:
        raise EnvironmentError(
            "Git is not available on this system. Please install Git and ensure it is in PATH."
        )

    self._notify("Updating regexes via Git...")

    with tempfile.TemporaryDirectory() as tmp:
        subprocess.run(
            [
                "git",
                "clone",
                "--depth",
                "1",
                "--filter=blob:none",
                "--sparse",
                "--branch",
                self.branch,
                self.repo_url,
                tmp,
            ],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        subprocess.run(
            [
                "git",
                "-C",
                tmp,
                "sparse-checkout",
                "set",
                self.sparse_dir,
                self.sparse_fixtures_dir,
                self.sparse_client_dir,
                self.sparse_device_dir,
            ],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
        )

        mapping = [
            (Path(tmp) / self.sparse_dir, self.upstream_path),
            (Path(tmp) / self.sparse_fixtures_dir, self.fixtures_upstream_path),
            (Path(tmp) / self.sparse_client_dir, self.client_upstream_dir),
            (Path(tmp) / self.sparse_device_dir, self.device_upstream_dir),
        ]

        for src, dst in mapping:
            if not src.exists():
                raise FileNotFoundError(src)
            self._clean_dir(dst)
            shutil.copytree(src, dst, dirs_exist_ok=True)
            (dst / "__init__.py").touch()

        clean_and_format_regexes()

    self._notify("Git update complete")


def _normalize_github_url(url: str):
    parsed = urlparse(url.strip())
    if parsed.netloc != "github.com":
        raise ValueError("Invalid GitHub URL")

    parts = parsed.path.strip("/").split("/")
    if len(parts) < 4 or parts[2] != "tree":
        raise ValueError("Expected GitHub tree URL")

    return {
        "owner": parts[0],
        "repo": parts[1],
        "branch": parts[3],
        "path": "/".join(parts[4:]),
    }


async def _check_rate_limit(session, token):
    headers = {"Authorization": f"token {token}"} if token else {}
    async with session.get("https://api.github.com/rate_limit", headers=headers) as r:
        data = await r.json()
        core = data.get("rate") or {}
        remaining = core.get("remaining", 0)
        if remaining < 5:
            reset = datetime.fromtimestamp(core.get("reset", 0), tz=timezone.utc)
            raise RuntimeError(f"Rate limit low ({remaining}), resets at {reset.isoformat()}")


async def _walk_contents(session, api_url, token, out):
    async with session.get(api_url) as r:
        r.raise_for_status()
        items = await r.json()

        if isinstance(items, dict):
            out.append(items)
            return

        for item in items:
            if item["type"] == "dir":
                await _walk_contents(session, item["url"], token, out)
            else:
                out.append(item)


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(),
    retry=retry_if_exception_type(Exception),
)
async def _download(session, url, path):
    async with session.get(url) as r:
        r.raise_for_status()
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(await r.read())


async def _api_download_tree(self, github_url, output_dir):
    meta = _normalize_github_url(github_url)

    api_url = (
        f"https://api.github.com/repos/{meta['owner']}/{meta['repo']}"
        f"/contents/{meta['path']}?ref={meta['branch']}"
    )

    headers = {"Authorization": f"token {self.github_token}"} if self.github_token else {}

    subtree_root = meta["path"].rstrip("/") + "/"

    async with aiohttp.ClientSession(headers=headers) as session:
        await _check_rate_limit(session, self.github_token)

        contents: List[dict] = []
        await _walk_contents(session, api_url, self.github_token, contents)

        sem = asyncio.Semaphore(10)

        async def bounded(item):
            async with sem:
                if not item.get("download_url"):
                    return

                relative_path = item["path"]
                if not relative_path.startswith(subtree_root):
                    raise RuntimeError(
                        f"Unexpected path {relative_path}, expected prefix {subtree_root}"
                    )

                relative_path = relative_path[len(subtree_root) :]

                await _download(
                    session,
                    item["download_url"],
                    Path(output_dir) / relative_path,
                )

        await asyncio.gather(*(bounded(i) for i in contents))


@register(UpdateMethod.API)
def _update_with_api(self: Regexes):
    self._notify("Updating regexes via GitHub API...")

    tasks = [
        (
            "https://github.com/matomo-org/device-detector/tree/master/regexes",
            self.upstream_path,
        ),
        (
            "https://github.com/matomo-org/device-detector/tree/master/Tests/fixtures",
            self.fixtures_upstream_path,
        ),
        (
            "https://github.com/matomo-org/device-detector/tree/master/Tests/Parser/Client/fixtures",
            self.client_upstream_dir,
        ),
        (
            "https://github.com/matomo-org/device-detector/tree/master/Tests/Parser/Device/fixtures",
            self.device_upstream_dir,
        ),
    ]

    async def runner():
        for url, path in tasks:
            self._clean_dir(path)
            await _api_download_tree(self, url, path)
            (path / "__init__.py").touch()

    asyncio.run(runner())

    clean_and_format_regexes()

    self._notify("API update complete")
